"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import os

# disable tf deprecation warnings
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
import tensorflow as tf

tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

import tifffile
from matplotlib import pyplot as plt
import skimage.exposure
import skimage.morphology
import numpy as np
import numexpr
from nuset import Nuset
from tqdm import tqdm
from typing import List, Tuple
from matplotlib import cm
from matplotlib.colors import LinearSegmentedColormap
from multiprocessing import Pool, cpu_count
from scipy.ndimage import label as label_image, generate_binary_structure
from joblib import Parallel, delayed
import scipy.sparse
from scipy.spatial import cKDTree, ConvexHull
from scipy.spatial.qhull import QhullError

from PyQt5 import QtWidgets, QtCore, QtGui
from qtap import Function
from ..core import ViewerUtils, ViewerWorkEnv
from ...pyqtgraphCore import GraphicsLayoutWidget, ImageItem, ViewBox
from ...pyqtgraphCore.console.Console import ConsoleWidget
from ...common.qdialogs import *
from ...common.configuration import console_history_path
from ...common.utils import HdfTools
import logging


logger = logging.getLogger()

cmap_magenta = LinearSegmentedColormap.from_list('magentas', [(0, 0, 0), (0,)*3, (1, 0, 1)])
cmap_cyan = LinearSegmentedColormap.from_list('cyans', [(0, 0, 0), (0,)*3, (0, 1, 1)])


def get_projection(img: np.ndarray, proj: str = 'std') -> np.ndarray:
    func = getattr(np, proj)

    return func(img, axis=2)


def get_preprocess(
        img: np.ndarray,
        do_preprocess: bool = True,
        do_sigmoid: bool = True,
        sigmoid_cutoff: float = 40.0,
        sigmoid_gain: float = 0.10,
        sigmoid_invert: bool = False,
        do_equalize: bool = False,
        equalize_lower: float = -0.1,
        equalize_upper: float = 1.0,
        equalize_kernel: int = 8,
) -> np.ndarray:
    """

    :param img: input image
    :param do_sigmoid: Perform sigmoid correction
    :param sigmoid_cutoff: sigmoid cutoff
    :param sigmoid_gain: sigmoid gain
    :param sigmoid_invert: invert after sigmoid correction (necessary sometimes)
    :param do_equalize: Perform histogram equalization
    :param equalize_kernel: Size of equalization kernel
    :return: preprocessed image
    """
    if not do_preprocess:
        return normalize_image(img, np.float32)

    img = img.astype(np.float32)

    if do_sigmoid:
        img = skimage.exposure.adjust_sigmoid(
            img,
            cutoff=sigmoid_cutoff,
            gain=sigmoid_gain,
            inv=sigmoid_invert
        )

    if do_equalize:
        # equalize_hist need floats that stay in the range of -1 to 1
        rng = (equalize_lower, equalize_upper)  # user specified range

        min_t = np.nanmin(img)
        max_t = np.nanmax(img)
        range_t = max_t - min_t
        nomin = (img - min_t) * (rng[1] - rng[0])
        img = nomin / range_t + rng[0]  # normalized img
        img = skimage.exposure.equalize_adapthist(img, kernel_size=equalize_kernel)

    # float32 is best for GPU I think?
    img = normalize_image(img, np.float32)

    return img


def wrap_preprocess(kwargs):
    return get_preprocess(**kwargs)


# from NuSeT processing cleanimage function
def get_postprocess(
        img: np.ndarray,
        do_postprocess: bool = False,
        abs_obj_threshold: int = -1,
        rel_obj_threshold: int = -1,
        obj_connectivity: int = 2,
        abs_hol_threshold: int = -1,
        rel_hol_threshold: int = 1,
        hol_connectivity: int = 2
    ) -> np.ndarray:

    if not do_postprocess:
        return img

    if not img.size > 0:
        return img

    image = img.astype(np.bool)
    im_label = skimage.morphology.label(image, connectivity=1)
    num_cells = np.max(im_label)
    mean_area = np.sum(image).astype(np.float32) / num_cells

    if rel_obj_threshold != -1:
        min_obj_size = mean_area / rel_obj_threshold
    elif abs_hol_threshold != -1:
        min_obj_size = abs_obj_threshold
    else:
        min_obj_size = 0
    image = skimage.morphology.remove_small_objects(image, min_size=min_obj_size, connectivity=obj_connectivity)

    if abs_hol_threshold == -1:
        min_hol_size = mean_area / rel_hol_threshold
    else:
        min_hol_size = abs_hol_threshold
    image = skimage.morphology.remove_small_holes(image, min_size=min_hol_size, connectivity=hol_connectivity)

    return image.astype(np.uint8)


# adapted from https://github.com/ver228/tierpsy-tracker/blob/master/tierpsy/analysis/compress/compressVideo.py
def normalize_image(img, dtype):
    # normalise image intensities between 0 and the max range of the dtype
    imax = (img.max() + np.abs(img.min())).astype(dtype)
    # we will set the values between 0 and max value of the float/int type
    imin = 0

    # max value of the dtype
    if np.issubdtype(dtype, np.floating):
        factor = (np.finfo(dtype).max - 1) / (imax - imin)
    elif np.issubdtype(dtype, np.integer):
        factor = (np.iinfo(dtype).max - 1) / (imax - imin)

    imgN = numexpr.evaluate('(img-imin)*factor')
    imgN = imgN.astype(dtype)

    return imgN


# adapted from skimage.morphology.remove_small_objects
def remove_large_objects(ar, max_size=128, connectivity=1):
    if max_size == -1:  # shortcut for efficiency
        return ar

    out = ar.copy()

    if out.dtype == bool:
        selem = generate_binary_structure(ar.ndim, connectivity)
        ccs = np.zeros_like(ar, dtype=np.int32)
        label_image(ar, selem, output=ccs)
    else:
        ccs = out

    try:
        component_sizes = np.bincount(ccs.ravel())
    except ValueError:
        raise ValueError("Negative value labels are not supported. Try "
                         "relabeling the input with `scipy.ndimage.label` or "
                         "`skimage.morphology.label`.")

    too_big = component_sizes > max_size
    too_big_mask = too_big[ccs]
    out[too_big_mask] = 0

    return out


def _get_sparse_mask(areas, i, edge_method, selem):
    temp = (areas[0] == i + 1)
    edge_func = getattr(skimage.morphology, edge_method)
    temp = edge_func(temp, selem=selem)

    return scipy.sparse.csr_matrix(temp.flatten('F'))


# adapted from https://github.com/flatironinstitute/CaImAn/blob/master/caiman/base/rois.py#L88
def get_sparse_masks(
        segmented_img: np.ndarray,
        raw_img_shape: tuple,
        edge_method: str,
        selem: int,
        transpose=True
    ) -> np.ndarray:
    # allocate array with same size as the raw input image
    # so that the dims match for CNMF
    img = np.zeros(raw_img_shape, dtype=segmented_img.dtype)

    # fill the allocated array with vals from the NuSeT segmentation
    # sometimes the segmented image has its dims trimmed by a few pixels
    if len(raw_img_shape) == 3:
        img[:, :segmented_img.shape[1], :segmented_img.shape[2]] = segmented_img
        struc = np.array(
            [[[0, 0, 0],
              [0, 0, 0],
              [0, 0, 0]],
             [[1, 1, 1],
              [1, 1, 1],
              [1, 1, 1]],
             [[0, 0, 0],
              [0, 0, 0],
              [0, 0, 0]]]
        )

    else:
        img[:segmented_img.shape[0], :segmented_img.shape[1]] = segmented_img
        struc = None

    areas = label_image(img, structure=struc)
    # A = np.zeros((np.prod(img.shape), areas[1]), dtype=bool)

    selem: np.array = np.ones((selem,)*len(segmented_img.shape))

    if areas[1] < 50:
        logger.info("Less than 50 regions, not parallelizing")
        sparses = []
        for i in tqdm(range(areas[1])):
            sparses.append(
                _get_sparse_mask(areas, i, edge_method, selem)
            )
    else:
        logger.info("Greater than 50 regions, parallelizing with joblib")
        sparses = Parallel(n_jobs=cpu_count(), verbose=5)(
            delayed(_get_sparse_mask)(areas, i, edge_method, selem) for i in range(areas[1])
        )

    if transpose:
        return scipy.sparse.vstack(sparses).T.toarray()
    else:
        return scipy.sparse.vstack(sparses).toarray()


def get_colored_mask(m: np.ndarray, shape: tuple):
    random_colors = np.random.choice(range(24), m.shape[1], replace=True)

    colored_mask = np.zeros(shape, dtype=np.uint16)

    for i in tqdm(range(m.shape[1])):
        colored_mask[m[:, i].reshape(shape)] = random_colors[i]

    return colored_mask


def area_to_vertices(a: np.ndarray) -> np.ndarray:
    """
    Get the vertices of the area in a binary mask
    :param a: binary mask
    :return: 2D array of x-y coordinates for all the vertices of the area in the mask
    """
    xs, ys = np.where(a)

    points = np.array((xs, ys)).T

    kdt = cKDTree(points)
    vertices = []

    for p in points:
        if len(kdt.query_ball_point(p, 1)) < 5:
            vertices.append(p)

    vs = np.vstack(vertices)

    return vs


def area_to_hull(a: np.ndarray) -> np.ndarray:
    """
    Get the vertices of a convex hull generated from an area represented by a binary mask
    :param a: binary mask
    :return: 2D array of x-y coordinates for the hull vertices
    """
    xs, ys = np.where(a)
    points = np.array((xs, ys)).T
    hull = ConvexHull(points, qhull_options='Qs')
    vs = points[hull.vertices]
    return vs


class ModuleGUI(QtWidgets.QMainWindow):
    def __init__(self, parent, viewer_ref):
        QtWidgets.QMainWindow.__init__(self)
        self.setWindowTitle('Nuset Segmentation')

        self.resize(1000, 900)
        self.menubar = self.menuBar()
        # self.setMenuBar(self.menubar)
        # self.menubar.setGeometry(QtCore.QRect(0, 0, 1000, 900))
        # self.menubar.setObjectName("menubar")


        self.widget = NusetWidget(self, viewer_ref)

        self.setCentralWidget(self.widget)

        self.dock_console = QtWidgets.QDockWidget(self)
        self.dock_console.setFeatures(
            QtWidgets.QDockWidget.DockWidgetFloatable | QtWidgets.QDockWidget.DockWidgetMovable
        )
        # self.dockWidgetContents = QtWidgets.QWidget()
        # self.dock_console.setWidget(self.dockWidgetContents)
        self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, self.dock_console)
        self.dock_console.hide()

        histfile = os.path.join(console_history_path, 'nuset_widget.pik')
        self.console_widget = ConsoleWidget(
            parent=self, namespace={'this': self.widget}, historyFile=histfile
        )
        self.dock_console.setWidget(self.console_widget)
        self.dock_console.setWindowTitle("Console")
        self.dock_console.hide()

        self.menu_export = self.menubar.addMenu('&Export')

        self.action_export_options = QtWidgets.QAction(text='Open Exporter')
        self.action_export_options.setToolTip(
            "Export as sparse matrix that can be used for \n"
            "seeding spatial components in Caiman CNMF"
        )
        self.menu_export.addAction(self.action_export_options)

        # self.action_export_cnmf_seeds = QtWidgets.QAction(text='CNMF seeds')
        # self.action_export_cnmf_seeds.setToolTip(
        #     "Export as sparse matrix that can be used for \n"
        #     "seeding spatial components in Caiman CNMF"
        # )
        # self.menu_export.addAction(self.action_export_cnmf_seeds)
        #
        # self.action_export_viewer = QtWidgets.QAction("Manual ROIs - Viewer")
        # self.menu_export.addAction(self.action_export_viewer)
        # self.action_export_viewer.setToolTip(
        #     "Export as Manual ROIs into the Viewer"
        # )

        self.menu_view = self.menubar.addMenu('&View')

        self.action_view_console = QtWidgets.QAction(text="Console")
        self.menu_view.addAction(self.action_view_console)
        self.action_view_console.setText("Console")
        self.action_view_console.setCheckable(True)
        self.action_view_console.setChecked(False)

        self.action_view_console.toggled.connect(
            lambda: self.dock_console.setVisible(
                self.action_view_console.isChecked()
            )
        )

        self.action_export_options.triggered.connect(self.widget.export_widget.show)


class ExportWidget(QtWidgets.QWidget):
    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self)
        self.resize(600, 850)
        self.setWindowTitle('Nuset Exporter')
        self.nuset_widget: NusetWidget = parent

        self.vlayout = QtWidgets.QVBoxLayout(self)

        hlayout_threshold = QtWidgets.QHBoxLayout(self)
        label_threshold = QtWidgets.QLabel(self)
        label_threshold.setText("separation threshold:")
        hlayout_threshold.addWidget(label_threshold)

        # TODO: I should really use qtap here
        self.spinbox_thr = QtWidgets.QDoubleSpinBox(self)
        self.spinbox_thr.setMaximum(1.0)
        self.spinbox_thr.setMinimum(0.0)
        self.spinbox_thr.setValue(0.5)
        self.spinbox_thr.setSingleStep(0.1)
        self.spinbox_thr.setDecimals(2)
        self.spinbox_thr.setToolTip(
            "Threshold for cell separation"
        )
        hlayout_threshold.addWidget(self.spinbox_thr)
        hlayout_threshold.addItem(
            QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        )
        self.vlayout.addLayout(hlayout_threshold)

        # sliders are overrated
        # self.slider_thr = QtWidgets.QSlider(self, orientation=QtCore.Qt.Horizontal)
        # self.slider_thr.setMaximum(100)
        # self.slider_thr.setMinimum(0)
        # self.slider_thr.setValue(50)
        # self.slider_thr.setToolTip(
        #     "Threshold for cell separation"
        # )
        # self.slider_thr.valueChanged.connect(
        #     lambda v: self.spinbox_thr.setValue(float(v) / 100.0)
        # )
        # self.vlayout.addWidget(self.slider_thr)
        #
        # self.spinbox_thr.valueChanged.connect(
        #     lambda v: self.slider_thr.setValue(int(v * 100))
        # )

        hlayout_minmax_size = QtWidgets.QHBoxLayout(self)

        label_minsize = QtWidgets.QLabel(self)
        label_minsize.setText('min size (pixels)')
        label_minsize.setToolTip(
            "Minimum size for accepting an object, in pixels"
        )
        hlayout_minmax_size.addWidget(label_minsize)

        self.spinbox_minsize = QtWidgets.QSpinBox(self)
        self.spinbox_minsize.setMinimum(0)
        self.spinbox_minsize.setMaximum(999)
        self.spinbox_minsize.setValue(1)
        self.spinbox_minsize.setToolTip(
            "Minimum size for accepting an object, in pixels"
        )
        hlayout_minmax_size.addWidget(self.spinbox_minsize)


        label_maxsize = QtWidgets.QLabel(self)
        label_maxsize.setText("max size (pixels)")
        label_maxsize.setToolTip(
            "Maximum size of accepting an object, in pixels"
        )
        hlayout_minmax_size.addWidget(label_maxsize)

        self.spinbox_maxsize = QtWidgets.QSpinBox(self)
        self.spinbox_maxsize.setMinimum(-1)
        self.spinbox_maxsize.setMaximum(999999)
        self.spinbox_maxsize.setValue(-1)
        self.spinbox_maxsize.setToolTip(
            "Maximum size of accepting an object, in pixels.\nSet as -1 to disable."
        )

        hlayout_minmax_size.addWidget(self.spinbox_maxsize)
        hlayout_minmax_size.addItem(
            QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        )
        self.vlayout.addLayout(hlayout_minmax_size)

        hlayout_conn_struc = QtWidgets.QHBoxLayout(self)

        label_connectivity = QtWidgets.QLabel(self)
        label_connectivity.setText("connectivity")
        hlayout_conn_struc.addWidget(label_connectivity)

        self.spinbox_connectivity = QtWidgets.QSpinBox(self)
        self.spinbox_connectivity.setMinimum(1)
        self.spinbox_connectivity.setMaximum(99)
        self.spinbox_connectivity.setValue(1)
        self.spinbox_connectivity.setToolTip(
            "Connectivity around each pixel"
        )
        hlayout_conn_struc.addWidget(self.spinbox_connectivity)

        label_selem = QtWidgets.QLabel(self)
        label_selem.setText("structural_element")
        hlayout_conn_struc.addWidget(label_selem)

        self.spinbox_selem = QtWidgets.QSpinBox(self)
        self.spinbox_selem.setMinimum(0)
        self.spinbox_selem.setMaximum(50)
        self.spinbox_selem.setValue(3)
        hlayout_conn_struc.addWidget(self.spinbox_selem)

        hlayout_conn_struc.addItem(
            QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        )
        self.vlayout.addLayout(hlayout_conn_struc)

        hlayout_edge_multi_2d = QtWidgets.QHBoxLayout(self)

        label_edge_method = QtWidgets.QLabel(self)
        label_edge_method.setText("Edge method")
        hlayout_edge_multi_2d.addWidget(label_edge_method)

        self.combo_edge_method = QtWidgets.QComboBox(self)
        self.combo_edge_method.addItems(
            ['closing', 'erosion', 'dilation']
        )
        self.combo_edge_method.setToolTip(
            "Method to deal with object edges"
        )
        hlayout_edge_multi_2d.addWidget(self.combo_edge_method)

        self.checkbox_multi2d = QtWidgets.QCheckBox(self)
        self.checkbox_multi2d.setText("Use multi-2D for 3D data")
        self.checkbox_multi2d.setToolTip(
            "Creates a separate sparse array for each z-plane, "
            "instead of a single sparse array for the entire stack."
        )
        hlayout_edge_multi_2d.addWidget(self.checkbox_multi2d)

        hlayout_edge_multi_2d.addItem(
            QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        )
        self.vlayout.addLayout(hlayout_edge_multi_2d)

        button_apply_thr_params = QtWidgets.QPushButton(self)
        button_apply_thr_params.setText("Apply")
        button_apply_thr_params.setStyleSheet("font-weight: bold")
        button_apply_thr_params.clicked.connect(
            lambda: self.apply_threshold()
        )
        self.vlayout.addWidget(button_apply_thr_params)

        hlayout_export_buttons = QtWidgets.QHBoxLayout(self)

        button_export_cnmf_2d = QtWidgets.QPushButton(self)
        button_export_cnmf_2d.setText("Export 2D CNMF Seeds")
        button_export_cnmf_2d.setStyleSheet("font-weight: bold")
        button_export_cnmf_2d.clicked.connect(
            lambda: self.save_masks_cnmf()
        )
        hlayout_export_buttons.addWidget(button_export_cnmf_2d)

        # button_export_cnmf_3d = QtWidgets.QPushButton(self)
        # button_export_cnmf_3d.setText("Export 3D CNMF Seeds")
        # button_export_cnmf_3d.setStyleSheet("font-weight: bold")
        # hlayout_export_buttons.addWidget(button_export_cnmf_3d)

        button_export_viewer_rois = QtWidgets.QPushButton(self)
        button_export_viewer_rois.setText("Export to Viewer as Manual ROIs")
        button_export_viewer_rois.setStyleSheet("font-weight: bold")
        button_export_viewer_rois.clicked.connect(
            lambda: self.export_to_viewer()
        )
        hlayout_export_buttons.addWidget(button_export_viewer_rois)

        self.vlayout.addLayout(hlayout_export_buttons)

        self.glw = GraphicsLayoutWidget(self)
        self.imgitem = ImageItem()
        self.viewbox = self.glw.addViewBox()
        self.viewbox.setAspectLocked(True)
        self.viewbox.addItem(self.imgitem)

        self.vlayout.addWidget(self.glw)

        self.colored_mask: np.ndarray = np.empty(0)
        self.masks: Union[np.ndarray, List[np.ndarray]] = np.empty(0)
        self.binary_shape: tuple = None

        colormap = cm.get_cmap("nipy_spectral")  # cm.get_cmap("CMRmap")
        colormap._init()
        lut = (colormap._lut * 255).view(np.ndarray)
        self.imgitem.setLookupTable(lut)

        self.nuset_widget.sig_zlevel_changed.connect(
            self.set_colored_mask
        )

        self.setLayout(self.vlayout)

    # TODO: I should really use qtap here
    def get_params(self):
        selem = self.spinbox_selem.value()

        return {
            'thr': self.spinbox_thr.value(),
            'selem': selem if selem > 0 else None,
            'edge_method': self.combo_edge_method.currentText(),
            'minsize': self.spinbox_minsize.value(),
            'maxsize': self.spinbox_maxsize.value(),
            'connectivity': self.spinbox_connectivity.value(),
            'multi-2d': self.checkbox_multi2d.isChecked()
        }

    def get_all_params(self):
        """
        All params, including those used for segmentation before the export GUI.
        :return:
        """
        d = \
            {
                'nuset_segment_params': self.nuset_widget.params_final,
                'nuset_export_params': self.get_params(),
            }

        return d

    def set_colored_mask(self, z):
        if self.colored_mask.size > 0:
            self.imgitem.setImage(
                self.colored_mask[:, :, z]
            )

    def _make_binary(self, seg_img, params, shape):
        binary_ = np.full(seg_img.shape, False, dtype=np.bool)
        binary_[seg_img > params['thr']] = True

        binary = np.full(shape, False, np.bool)

        if len(shape) > 2:
            binary[:binary_.shape[0], :binary_.shape[1], :binary_.shape[2]] = binary_
        else:
            binary[:binary_.shape[0], :binary_.shape[1]] = binary_

        if params['minsize'] > 0:
            logger.info("Removing small objs")
            binary = skimage.morphology.remove_small_objects(
                binary, min_size=params['minsize'], connectivity=params['connectivity']
            )
            binary = remove_large_objects(
                binary, params['maxsize'], connectivity=params['connectivity']
            )

        return binary

    def _apply_threshold_2d(self):
        params = self.get_params()
        shape = self.nuset_widget.imgs_projected[0].T.shape
        seg_img = self.nuset_widget.imgs_postprocessed[0].T

        logger.info("Thresholding")
        binary = self._make_binary(seg_img, params, shape)#.T
        self.binary_shape = binary.shape

        logger.info("Creating sparse masks")
        self.masks = get_sparse_masks(
            segmented_img=binary,
            raw_img_shape=shape,
            edge_method=params['edge_method'],
            selem=params['selem'],
        )

        self.colored_mask = get_colored_mask(self.masks, binary.T.shape)
        self.imgitem.setImage(self.colored_mask)

    def _apply_threshold_3d(self):
        params = self.get_params()
        # 3D
        seg_img = np.stack(self.nuset_widget.imgs_postprocessed)
        shape = np.stack(self.nuset_widget.imgs_projected).shape

        logger.info("Thresholding")
        # binary array
        binary = self._make_binary(seg_img, params, shape)
        self.binary_shape = binary.shape

        logger.info("Creating sparse masks")
        self.masks = get_sparse_masks(
            segmented_img=binary,
            raw_img_shape=shape,
            edge_method=params['edge_method'],
            selem=params['selem']
        )

        self.colored_mask = get_colored_mask(self.masks, binary.T.shape)
        self.imgitem.setImage(
            self.colored_mask[:, :, self.nuset_widget.zlevel]
        )

    def _apply_threshold_2d_multi(self):
        params = self.get_params()

        shape = self.nuset_widget.imgs_projected[0].T.shape

        masks = []
        logger.info("Thresholding & Creating sparse masks")
        for ix in tqdm(range(len(self.nuset_widget.imgs_postprocessed))):
            seg_img = self.nuset_widget.imgs_postprocessed[ix].T

            binary = self._make_binary(seg_img, params, shape)
            self.binary_shape = binary.shape

            masks.append(
                get_sparse_masks(
                    segmented_img=binary,
                    raw_img_shape=shape,
                    edge_method=params['edge_method'],
                    selem=params['selem']
                )
            )

        self.masks = masks

        logger.info("Coloring masks")
        colored_masks = []
        for ix in tqdm(range(len(self.masks))):
            colored_masks.append(
                get_colored_mask(
                    self.masks[ix], binary.shape
                )
            )

        self.colored_mask = np.stack(colored_masks).T
        self.imgitem.setImage(self.colored_mask[:, :, self.nuset_widget.zlevel])

    @present_exceptions()
    def apply_threshold(self):
        if not self.nuset_widget.imgs_postprocessed:
            raise ValueError("You must segment images before you can proceed.")

        for img in self.nuset_widget.imgs_postprocessed:
            if not img.size > 0:
                raise ValueError("Your must segment the entire stack before you can proceed.")

        if self.nuset_widget.z_max > 0:
            if QtWidgets.QMessageBox.question(
                    self, 'Export for CNMF?', 'This may take a few minutes, and could take ~10 minutes '
                                              'if segmenting a large stack. Proceed?',
            ) == QtWidgets.QMessageBox.No:
                return

        # 3D
        if self.nuset_widget.z_max > 0:
            if self.checkbox_multi2d.isChecked():
                self._apply_threshold_2d_multi()
            else:
                self._apply_threshold_3d()
        # 2D
        else:
            self._apply_threshold_2d()

    def _check_save_masks(self):
        if not self.nuset_widget.imgs_postprocessed:
            raise ValueError("You must segment images before you can proceed.")

        for img in self.nuset_widget.imgs_postprocessed:
            if not img.size > 0:
                raise ValueError("Your must segment the entire stack before you can proceed.")

        if not len(self.masks) > 0:
            raise ValueError("You must apply the threshold before you can proceed")

    @use_save_file_dialog('Save masks', None, '.h5')
    @present_exceptions()
    def save_masks_cnmf(self, path):
        logger.info("Please wait, saving masks...")
        if self.get_params()['multi-2d']:
            sm = {}
            for z in range(len(self.masks)):
                sm[str(z)] = self.masks[z]  # each plane is addressed as an hdf5 group
        else:
            sm = self.masks

        d = \
            {
                'sparse_mask': sm,
                'segment_params': self.get_all_params()
            }
        HdfTools.save_dict(d, path, 'data')
        logger.info("Finishing saving masks!")

    @present_exceptions()
    def export_to_viewer(self):
        if not self.nuset_widget.imgs_postprocessed:
            raise ValueError("You must segment images before you can proceed.")

        for img in self.nuset_widget.imgs_postprocessed:
            if not img.size > 0:
                raise ValueError("Your must segment the entire stack before you can proceed.")

        if not self.masks.size > 0:
            raise ValueError("You must apply the threshold before you can proceed")

        roi_manager = self.nuset_widget.vi.viewer.parent().get_module('roi_manager')
        roi_manager.start_backend('Manual')

        if QtWidgets.QMessageBox.question(
            self, 'Use Convex Hull?',
            'Use a Convex Hull to get the vertices? [Yes]\n'
            'This is recommended if you are importing as ManualROIs.\n'
            'If you select [No] ALL vertices for the masks will be imported and '
            'this could be extremely slow!',
            QtWidgets.QMessageBox.No, QtWidgets.QMessageBox.Yes
        ) == QtWidgets.QMessageBox.Yes:
            method = 'ConvexHull'
        else:
            method = 'cKDTree'

        if self.nuset_widget.z_max > 0:
            shape = self.binary_shape
        else:
            shape = self.binary_shape[::-1]

        if method == 'ConvexHull':
            self._export_hulls(shape)
        elif method == 'cKDTree':
            self._export_ckdtrees(shape)

        self.nuset_widget.vi.viewer.workEnv.history_trace.append(
            {
                'nuset_segmentation': \
                    {
                        **self.get_all_params(),
                        'vertex_method': method
                    }
            }
        )

    def _export_hulls(self, shape):
        for i in tqdm(range(self.masks.shape[1])):
            try:
                vs = area_to_hull(
                    self.masks[:, i].reshape(shape)
                )
            except QhullError:
                logger.info(f"Skipping {i}, not enough points for convex hull")
            else:
                self.nuset_widget.vi.viewer.workEnv.roi_manager.add_roi_from_points(
                    xs=vs[:, 0], ys=vs[:, 1]
                )

    def _export_ckdtrees(self, shape):
        for i in tqdm(range(self.masks.shape[1])):
            vs = area_to_vertices(
                self.masks[:, i].reshape(shape)
            )

            self.nuset_widget.vi.viewer.workEnv.roi_manager.add_roi_from_points(
                xs=vs[:, 0], ys=vs[:, 1]
            )


class NusetWidget(QtWidgets.QWidget):
    sig_zlevel_changed = QtCore.pyqtSignal(int)

    def __init__(self, parent, viewer_ref):
        QtWidgets.QWidget.__init__(self, parent)

        self.vlayout = QtWidgets.QVBoxLayout(self)

        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        self.vlayout.addWidget(self.splitter)

        self.left_layout = QtWidgets.QVBoxLayout(self)

        self.left_widget = QtWidgets.QWidget(self.splitter)
        self.left_widget.setLayout(self.left_layout)

        hlayout_custom_img = QtWidgets.QHBoxLayout(self.left_widget)
        self.lineedit_custom_img_path = QtWidgets.QLineEdit(self.left_widget)
        self.lineedit_custom_img_path.setPlaceholderText("Path to custom image or stack")
        self.lineedit_custom_img_path.setToolTip(
            "Use a different image or stack for segmentation, such as an image or stack of nuclei."
        )
        self.button_use_custom = QtWidgets.QPushButton(self.left_widget)
        self.button_use_custom.setMaximumWidth(50)
        self.button_use_custom.setText("Set")
        hlayout_custom_img.addWidget(self.lineedit_custom_img_path)
        hlayout_custom_img.addWidget(self.button_use_custom)
        self.left_layout.addLayout(hlayout_custom_img)

        hlayout_projection_radios = QtWidgets.QHBoxLayout(self.left_widget)
        self.radio_std = QtWidgets.QRadioButton(self.left_widget)
        self.radio_std.setText("Std. Deviation")
        self.radio_max = QtWidgets.QRadioButton(self.left_widget)
        self.radio_max.setText("Max")
        self.radio_mean = QtWidgets.QRadioButton(self.left_widget)
        self.radio_mean.setText("Mean")

        self.projection_option: str = ''

        for w in [self.radio_std, self.radio_max, self.radio_mean]:
            hlayout_projection_radios.addWidget(w)
            w.clicked.connect(self.update_projection)

        label_projections = QtWidgets.QLabel(self.left_widget)
        label_projections.setText("Choose Input Projection")
        label_projections.setStyleSheet("font-weight: bold")

        self.left_layout.addWidget(label_projections)
        self.left_layout.addLayout(hlayout_projection_radios)

        # projection image
        self.glw_raw = GraphicsLayoutWidget(self.left_widget)
        self.imgitem_raw = ImageItem()
        self.viewbox_raw = self.glw_raw.addViewBox()
        self.viewbox_raw.setAspectLocked(True)
        self.viewbox_raw.addItem(self.imgitem_raw)

        self.left_layout.addWidget(self.glw_raw)

        hlayout_preprocess = QtWidgets.QHBoxLayout(self.left_widget)

        arg_opts_preprocess = \
            {
                'img': {'ignore': True},

                'sigmoid_cutoff': {
                    'step': 10,
                    'minmax': (0, 9999.9)
                },

                'sigmoid_gain': {
                    'step': 0.01
                },
                'equalize_lower': {
                    'step': 0.01,
                    'minmax': (-1.0, 1.0)
                },
                'equalize_upper': {
                    'step': 0.01,
                    'minmax': (-1.0, 1.0)
                }
            }

        self.preprocess_controls = \
            Function(
                get_preprocess,
                arg_opts=arg_opts_preprocess,
                parent=self.left_widget
            )
        self.preprocess_controls.button_set.hide()
        self.preprocess_controls.sig_changed.connect(self.update_preprocess)
        self.preprocess_controls.set_title("Pre-process Settings")

        hlayout_preprocess.addWidget(self.preprocess_controls.widget)

        self.zslider = QtWidgets.QSlider(parent=self.left_widget, orientation=QtCore.Qt.Vertical)
        self.zslider.setInvertedAppearance(True)
        self.zslider.setSingleStep(1)
        self.zslider.valueChanged.connect(self.update_zlevel)

        label_zslider = QtWidgets.QLabel(self.left_widget)
        label_zslider.setText("z-level")

        self.spinbox_zlevel = QtWidgets.QSpinBox(self.left_widget)
        self.spinbox_zlevel.setSingleStep(1)
        self.spinbox_zlevel.valueChanged.connect(self.zslider.setValue)
        self.zslider.valueChanged.connect(self.spinbox_zlevel.setValue)


        hlayout_preprocess.addWidget(label_zslider)
        hlayout_preprocess.addWidget(self.zslider)
        hlayout_preprocess.addWidget(self.spinbox_zlevel)

        self.left_layout.addLayout(hlayout_preprocess)

        # preprocess image item
        self.glw_preprocess = GraphicsLayoutWidget(self.left_widget)
        self.imgitem_preprocess = ImageItem()
        self.viewbox_preprocess = self.glw_preprocess.addViewBox()
        self.viewbox_preprocess.setAspectLocked(True)
        self.viewbox_preprocess.addItem(self.imgitem_preprocess)

        self.left_layout.addWidget(self.glw_preprocess)

        self.splitter.addWidget(self.left_widget)

        # segmentation stuff
        self.nuset_model = Nuset()

        self.right_layout = QtWidgets.QVBoxLayout(self)

        self.right_widget = QtWidgets.QWidget(self.splitter)
        self.right_widget.setLayout(self.right_layout)

        hlayout_nuset = QtWidgets.QHBoxLayout(self.right_widget)

        arg_opts_nuset = \
            {
                'image': {'ignore': True},

                'min_score': {
                    'minmax': (0.0, 1.0),
                    'step': 0.05
                },

                'nms_threshold': {
                    'minmax': (0.0, 1.0),
                    'step': 0.05
                },

                'rescale_ratio': {
                    'minmax': (0.1, 10.0),
                    'step': 0.25,
                    'val': 1.25  # not working if set to 1.0 with pyqtgraph ImageItem for some reason
                }

            }

        self.nuset_controls = \
            Function(
                self.nuset_model.predict,
                arg_opts=arg_opts_nuset,
                parent=self.right_widget
            )

        # self.nuset_controls.button_set.hide()
        self.nuset_controls.sig_set_clicked.connect(self.update_segmentation)
        self.nuset_controls.set_title("NuSeT Parameters")
        hlayout_nuset.addWidget(self.nuset_controls.widget)

        arg_opts_postprocess = \
            {
                'img': {'ignore': True},

                'abs_obj_threshold': {
                    'minmax': (-1, 100),
                    'step': 2,
                },

                'rel_obj_threshold': {
                    'minmax': (0, 100),
                    'step': 2
                },

                'obj_connectivity': {
                    'minmax': (0, 10),
                    'step': 1
                },

                'abs_hol_threshold': {
                    'minmax': (-1, 100),
                    'step': 2,
                },

                'rel_hol_threshold': {
                    'minmax': (0, 100),
                    'step': 2
                },

                'hol_connectivity': {
                    'minmax': (0, 100),
                    'step': 1
                }
            }

        self.postprocess_controls = \
            Function(
                get_postprocess,
                arg_opts=arg_opts_postprocess,
                parent=self.right_widget
            )
        self.postprocess_controls.button_set.hide()
        self.postprocess_controls.sig_changed.connect(self.update_postprocess)
        self.postprocess_controls.set_title("Post-process Parameters")

        hlayout_nuset.addWidget(self.postprocess_controls.widget)
        self.right_layout.addLayout(hlayout_nuset)

        self.checkbox_segment_current_plane = QtWidgets.QCheckBox(self.nuset_controls.widget)
        self.checkbox_segment_current_plane.setText("Only segment current plane")
        self.checkbox_segment_current_plane.setToolTip(
            "Performs segmentation for only the current plane, "
            "useful for trying out parameters."
        )
        self.checkbox_segment_current_plane.setChecked(True)
        self.nuset_controls.vlayout.addWidget(self.checkbox_segment_current_plane)

        self.glw_segmented = GraphicsLayoutWidget(self.right_widget)
        self.imgitem_segmented = ImageItem()
        self.imgitem_segmented_underlay = ImageItem()
        self.viewbox_segmented = self.glw_segmented.addViewBox()
        self.viewbox_segmented.setAspectLocked(True)
        self.viewbox_segmented.addItem(self.imgitem_segmented)
        self.viewbox_segmented.addItem(self.imgitem_segmented_underlay)

        # colormaps for the mask overlay plot
        cm_cyan = cm.get_cmap(cmap_cyan)
        cm_cyan._init()
        lut_cyan = (cm_cyan._lut * 255).view(np.ndarray)[128:-5]
        self.imgitem_segmented_underlay.setLookupTable(lut_cyan)

        # colormaps for the mask overlay plot
        cm_magenta = cm.get_cmap(cmap_magenta)
        cm_magenta._init()
        lut_magenta = (cm_magenta._lut * 255).view(np.ndarray)[:-5]  # so it displays as a mask
        self.imgitem_segmented.setLookupTable(lut_magenta)

        # allow transparency
        # self.imgitem_segmented.setCompositionMode(QtGui.QPainter.CompositionMode_Overlay)
        self.imgitem_segmented_underlay.setCompositionMode(QtGui.QPainter.CompositionMode_SourceOver)

        self.imgitem_segmented.setZValue(2)
        self.imgitem_segmented.setOpacity(0.5)
        self.imgitem_segmented_underlay.setZValue(1)
        self.imgitem_segmented_underlay.setOpacity(0.3)

        self.slider_underlay_opacity = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider_underlay_opacity.setMinimum(0)
        self.slider_underlay_opacity.setMaximum(100)
        self.slider_underlay_opacity.setValue(30)
        self.slider_underlay_opacity.valueChanged.connect(
            lambda v: self.imgitem_segmented_underlay.setOpacity(v / 100)
        )
        label_underlay_opacity = QtWidgets.QLabel(self.nuset_controls.widget)
        label_underlay_opacity.setText("Underlay opacity")
        self.nuset_controls.vlayout.addWidget(label_underlay_opacity)
        self.nuset_controls.vlayout.addWidget(self.slider_underlay_opacity)

        self.slider_mask_opacity = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider_mask_opacity.setMinimum(0)
        self.slider_mask_opacity.setMaximum(100)
        self.slider_mask_opacity.setValue(50)
        self.slider_mask_opacity .valueChanged.connect(
            lambda v: self.imgitem_segmented.setOpacity(v / 100)
        )
        label_segmented_opacity = QtWidgets.QLabel(self.nuset_controls.widget)
        label_segmented_opacity.setText("Segmentation opacity")
        self.nuset_controls.vlayout.addWidget(label_segmented_opacity)
        self.nuset_controls.vlayout.addWidget(self.slider_mask_opacity)

        self.combo_blend_mode = QtWidgets.QComboBox(self.nuset_controls.widget)
        self.combo_blend_mode.addItems(
            [
                'SourceOver',
                'Overlay',
                'Plus',
                'Multiply'
            ]
        )
        self.combo_blend_mode.setCurrentIndex(0)
        self.combo_blend_mode.currentTextChanged.connect(
            lambda opt: self.imgitem_segmented.setCompositionMode(
                getattr(QtGui.QPainter, f'CompositionMode_{opt}')
            )
        )
        label_blend_mode = QtWidgets.QLabel(self.nuset_controls.widget)
        label_blend_mode.setText('Blend mode:')
        self.nuset_controls.vlayout.addWidget(label_blend_mode)
        self.nuset_controls.vlayout.addWidget(self.combo_blend_mode)

        self.right_layout.addWidget(self.glw_segmented)
        self.splitter.addWidget(self.right_widget)

        self.export_widget = ExportWidget(self)

        self.imgs_projected: List[np.ndarray] = []
        self.imgs_preprocessed: List[np.ndarray] = []
        self.imgs_segmented: List[np.ndarray] = []
        self.imgs_postprocessed: List[np.ndarray] = []

        self.input_img: np.ndarray = np.empty(0)
        self.zlevel = 0
        self.z_max: int = 0

        self.process_pool = Pool(cpu_count())

        self.error_label = QtWidgets.QLabel(self)
        self.error_label.setMaximumHeight(20)
        self.error_label.setStyleSheet("font-weight: bold; color: red")
        self.vlayout.addWidget(self.error_label)

        self.projection_option = ''
        self.params_final = None

        if viewer_ref is None:
            logger.info("Assuming testing mode")
        else:
            self.vi = ViewerUtils(viewer_ref)
            self.vi.viewer.sig_workEnv_changed.connect(self.set_input)
            self.set_input(self.vi.viewer.workEnv)

    @use_open_file_dialog('Open image file', None, ['*.tiff', '*.tif'])
    @present_exceptions('Error', 'Error loading custom input image')
    def set_custom_input(self, path, *args):
        self.clear_widget()
        img = tifffile.imread(path)
        self._set_input_image(img.T)

    def set_input(self, workEnv: ViewerWorkEnv):
        self.clear_widget()

        try:
            img = workEnv.imgdata._seq
        except:
            return

        self._set_input_image(img)

    def _set_input_image(self, img: np.ndarray):
        self.input_img = img

        if self.input_img.ndim == 4:
            self.z_max = self.input_img.shape[3] - 1
        else:
            self.z_max = 0

        self.zslider.valueChanged.disconnect(self.update_zlevel)
        self.zslider.setValue(0)
        self.zslider.setMaximum(self.z_max)
        self.spinbox_zlevel.setMaximum(self.z_max)
        self.zslider.valueChanged.connect(self.update_zlevel)

    def update_zlevel(self, z: int):
        self.zlevel = z
        self.sig_zlevel_changed.emit(z)

        for imgitem, imglist in zip(
            [
                self.imgitem_raw,
                self.imgitem_preprocess,
                self.imgitem_segmented_underlay,
                self.imgitem_segmented
            ],
            [
                self.imgs_projected,
                self.imgs_preprocessed,
                self.imgs_preprocessed,
                self.imgs_postprocessed
            ]
        ):
            if imglist:  # set if the list is not empty
                if imglist[z].size > 0:
                    imgitem.setImage(imglist[z])

    def update_projection(self):
        if not self.input_img.size > 0:
            self.error_label.setText("No input image")
            return

        if self.radio_std.isChecked():
            opt = 'std'
        elif self.radio_max.isChecked():
            opt = 'max'
        elif self.radio_mean.isChecked():
            opt = 'mean'

        if self.projection_option == opt:
            return
        
        self.projection_option = opt

        func = getattr(np, f'nan{opt}')

        logger.info("Updating Projection(s)")
        if self.input_img.ndim == 4:
            self.imgs_projected = [
                func(self.input_img[:, :, :, z], axis=2) for z in tqdm(range(self.z_max + 1))
            ]
        else:
            self.imgs_projected = [func(self.input_img, axis=2)]

        self.imgitem_raw.setImage(self.imgs_projected[self.zlevel])
        self.error_label.clear()

    def update_preprocess(self, params):
        if not self.imgs_projected:
            self.error_label.setText("Projection Image is Empty")
            return

        logger.info("Preprocessing Image(s)")

        # self.image_preprocesses = [
        #     get_preprocess(p, **params) for p in tqdm(self.image_projections)
        # ]

        kwargs = [{'img': img, **params} for img in self.imgs_projected]

        self.imgs_preprocessed = self.process_pool.map(wrap_preprocess, kwargs)

        self.imgitem_preprocess.setImage(self.imgs_preprocessed[self.zlevel])
        self.imgitem_segmented_underlay.setImage(self.imgs_preprocessed[self.zlevel])
        self.error_label.clear()

    def update_segmentation(self, params):
        if not self.imgs_preprocessed:
            self.error_label.setText("Preprocess Image is Empty")
            return

        logger.info("Segmenting Image(s)")
        if not self.checkbox_segment_current_plane.isChecked():
            self.imgs_segmented = [
                self.nuset_model.predict(img, **params) for img in tqdm(self.imgs_preprocessed)
            ]
        else:
            self.imgs_segmented = [
                np.empty(0) for i in range(self.z_max + 1)
            ]

            self.imgs_segmented[self.zlevel] = self.nuset_model.predict(
                self.imgs_preprocessed[self.zlevel], **params
            )

        # postprocess funciton will just return the segmented img if do_postprocess is False
        self.update_postprocess(self.postprocess_controls.get_data())

        self.params_final = \
            {
                'method': 'NuSeT',
                'projection_option': self.projection_option,
                **self.preprocess_controls.get_data(),
                **self.nuset_controls.get_data(),
                **self.postprocess_controls.get_data(),
            }

        self.error_label.clear()

    def update_postprocess(self, params):
        if not self.imgs_segmented:
            self.error_label.setText("Segmented Image is Empty")
            return

        logger.info("Postprocessing Image(s)")
        self.imgs_postprocessed = [
            get_postprocess(img, **params) for img in tqdm(self.imgs_segmented)
        ]

        self.imgitem_segmented.setImage(self.imgs_postprocessed[self.zlevel])
        self.error_label.clear()

    def clear_widget(self):
        self.imgs_projected.clear()
        self.imgs_preprocessed.clear()
        self.imgs_segmented.clear()
        self.imgs_postprocessed.clear()

        self.imgitem_raw.clear()
        self.imgitem_preprocess.clear()
        self.imgitem_segmented.clear()
        self.imgitem_segmented_underlay.clear()

        self.radio_std.setChecked(False)
        self.radio_max.setChecked(False)
        self.radio_mean.setChecked(False)

        self.projection_option = ''



from functools import partial

import numpy as np
from PyQt5 import QtWidgets, QtGui
from matplotlib import cm as matplotlib_color_map

from .... import pyqtgraphCore as pg
from ....viewer.core.common import ViewerUtils
from ....viewer.modules.roi_manager_modules.roi_types import ManualROI, ScatterROI, VolMultiCNMFROI
from ....common import configuration, get_project_manager
from typing import Union


class ROIList(list):
    """A list for holding ROIs of one type"""
    def __init__(self, ui, roi_types: type, viewer_interface: ViewerUtils):
        """
        Instantiate

        :param ui: The ui from the parent ModuleGUI, used to interact with the ROI list widget etc.
        :param roi_types: The type of ROI that this list will hold
        :param viewer_interface: ViewerUtils instance for interacting with the parent Viewer

        """
        super(ROIList, self).__init__()

        assert isinstance(ui.listWidgetROIs, QtWidgets.QListWidget)
        self.list_widget = ui.listWidgetROIs  #: ROI list widget
        self.list_widget.clear()
        self.list_widget.currentRowChanged.connect(self.set_current_index)

        # self.action_delete_roi = QtWidgets.QWidgetAction(ui.dockWidgetContents)
        # self.action_delete_roi.setText('Delete')
        # self.action_delete_roi.triggered.connect(partial(self.__delitem__, None))
        #
        # self._list_widget_context_menu = QtWidgets.QMenu(ui.dockWidgetContents)
        # self._list_widget_context_menu.addAction(self.action_delete_roi)
        #
        # self.list_widget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        # self.list_widget.customContextMenuRequested.connect(self.bulid_delete_menu)

        assert isinstance(ui.listWidgetROITags, QtWidgets.QListWidget)
        self.list_widget_tags = ui.listWidgetROITags  #: Tags list widget
        self.list_widget_tags.clear()

        self.roi_types = roi_types

        assert isinstance(ui.btnPlot, QtWidgets.QPushButton)
        self.btn_plot = ui.btnPlot
        self.btn_plot.clicked.connect(self.plot_manual_roi_regions)

        assert isinstance(ui.checkBoxLivePlot, QtWidgets.QCheckBox)
        self.live_plot_checkbox = ui.checkBoxLivePlot
        self.live_plot_checkbox.setChecked(False)
        if issubclass(self.roi_types, ManualROI):
            self.live_plot_checkbox.setEnabled(True)
        else:
            self.live_plot_checkbox.setEnabled(False)

        assert isinstance(ui.checkBoxShowAll, QtWidgets.QCheckBox)
        self.show_all_checkbox = ui.checkBoxShowAll
        self.show_all_checkbox.setChecked(True)
        self.show_all_checkbox.toggled.connect(self.slot_show_all_checkbox_clicked)

        assert isinstance(ui.btnSetROITag, QtWidgets.QPushButton)
        self.btn_set_tag = ui.btnSetROITag
        self.btn_set_tag.clicked.connect(self.slot_btn_set_tag)

        assert isinstance(ui.lineEditROITag, QtWidgets.QLineEdit)
        self.line_edit_tag = ui.lineEditROITag

        self.vi = viewer_interface  #: ViewrUtils instance

        self.current_index = -1  #: Current index (int)
        self.previous_index = -1  #: Previous index (int)

        get_project_manager().signal_project_config_changed.connect(self.update_roi_defs_from_configuration)

        # configuration.proj_cfg_changed.register(self.update_roi_defs_from_configuration)

    def append(
            self,
            roi: Union[ScatterROI, ManualROI, VolMultiCNMFROI],
            add_to_list_widget: bool = True,
    ):
        """Add an ROI instance to the list"""
        roi.add_to_viewer()

        roi_graphics_object = roi.get_roi_graphics_object()

        # If it's a pyqtgraph based ROI (ManualROI)
        if isinstance(roi_graphics_object, tuple(pg.ROI.__subclasses__())):
            roi_graphics_object.sigHoverEvent.connect(partial(self.highlight_roi, roi))
            roi_graphics_object.sigHoverEnd.connect(roi.reset_color)
            roi_graphics_object.sigRemoveRequested.connect(partial(self.__delitem__, roi))
            roi_graphics_object.sigRegionChanged.connect(partial(self._live_update_requested, roi))

        if add_to_list_widget:
            self.list_widget.addItem(str(self.__len__()))
            
        super(ROIList, self).append(roi)

    # def clear_all(self):
    #     self.list_widget.clear()
    #     self.list_widget_tags.clear()
    #     for ix in range(self.__len__()):
    #         roi = self.__getitem__(ix)
    #         roi.remove_from_viewer()

    # def set_pg_roi_graphics_state(self, ix, state):
    #     try:
    #         roi = self.__getitem__(ix)
    #     except IndexError:
    #         return
    #     pg_roi = roi.get_roi_graphics_object()
    #     pg_roi.setState(state)

    def clear_(self):
        """Cleanup of the list"""
        self.list_widget.clear()
        self.list_widget_tags.clear()
        self.disconnect_all()
        self._clear()
        # for i in range(self.__len__()):
        #     self.vi.viewer.status_bar_label.showMessage('Removing ROIs #:' + str(i))
        #     self.__delitem__(0)

    def _clear(self):
        for key in range(self.__len__()):
            self.vi.viewer.status_bar_label.showMessage('Removing ROIs #:' + str(key))
            self.vi.workEnv_changed('ROI Removed')
            roi = self.__getitem__(0)
            # bad workaround for now
            if not isinstance(roi, VolMultiCNMFROI):
                roi.remove_from_viewer()
            super(ROIList, self).__delitem__(0)
            if self.__len__() == 0:
                self.list_widget.clear()
                self.list_widget_tags.clear()
                return

    def __delitem__(self, key):
        """Delete an ROI from the list and cleanup from the viewer, reindex the colors etc."""
        if isinstance(key, ManualROI):
            key = self.index(key)
        elif key is None:
            key = self.current_index
        if key == -1:
            return
        self.vi.workEnv_changed('ROI Removed')
        roi = self.__getitem__(key)
        roi.remove_from_viewer()
        super(ROIList, self).__delitem__(key)
        if self.__len__() == 0:
            self.list_widget.clear()
            self.list_widget_tags.clear()
            return

        self.list_widget.takeItem(key)

        if not isinstance(roi, VolMultiCNMFROI):
            self._reindex_list_widget()
            self.reindex_colormap()

        if self.list_widget.count() > 0:
            self.list_widget.setCurrentRow(0)


        # self.list_widget.setCurrentRow(-1)

            # self.list_widget.setCurrentRow(min(0, key - 1))
            # self._reindex_list_widget()
            # self.reindex_colormap()
            # self.set_list_widget_tags()

    # def __del__(self):
    #     pass

    def disconnect_all(self):
        """Disconnect signals from the parent GUI"""
        self.list_widget.currentRowChanged.disconnect(self.set_current_index)
        # self.list_widget.customContextMenuRequested.disconnect(self.bulid_delete_menu)
        # self.action_delete_roi.triggered.disconnect(self.__delitem__)
        self.btn_plot.clicked.disconnect(self.plot_manual_roi_regions)
        self.show_all_checkbox.toggled.disconnect(self.slot_show_all_checkbox_clicked)
        self.btn_set_tag.clicked.disconnect(self.slot_btn_set_tag)

    def _reindex_list_widget(self):
        """Reindex ROI list"""
        for i in range(self.list_widget.count()):
            self.list_widget.item(i).setText(str(i))

    def reindex_colormap(self, random_shuffle=False):
        """Reindex the colors so they sequentially follow the HSV colormap"""
        cm = matplotlib_color_map.get_cmap('hsv')
        cm._init()
        lut = (cm._lut * 255).view(np.ndarray)

        cm_ixs = np.linspace(0, 210, self.__len__(), dtype=int)

        ixs = np.arange(self.__len__())
        color_ixs = np.copy(ixs)

        if random_shuffle:
            np.random.shuffle(color_ixs)

        for ix, color_ix in zip(ixs, color_ixs):
            c = lut[cm_ixs[color_ix]]
            try:
                roi = self.__getitem__(ix)
            except IndexError:
                return
            item = self.list_widget.item(ix)
            if item is not None:
                item.setBackground(QtGui.QBrush(pg.mkBrush(c)))

            roi.set_original_color(c)
            # roi.set_color(c)

    def __getitem__(self, item) -> Union[ManualROI, ScatterROI]:
        """Get an item (ROI) from the list"""
        return super(ROIList, self).__getitem__(item)

    def set_current_index(self, ix: int):
        """Set the current index"""
        if ix < 0:
            return
        if ix > self.__len__():
            return
        try:
            self.__getitem__(ix)
        except IndexError:
            return
        self.current_index = ix
        self.highlight_curve(ix)
        self.set_previous_index()
        if not self.show_all_checkbox.isChecked():
            self._hide_all_graphics_objects()
        self._show_graphics_object(ix)
        self.set_list_widget_tags()

    def highlight_roi(self, roi: Union[ManualROI, ScatterROI]):
        """Highlight an ROI in white, both the spatial visualization and the curve"""
        ix = self.index(roi)
        self.highlight_curve(ix)
        self.list_widget.setCurrentRow(ix)

    def highlight_curve(self, ix: int):
        """Highlight the curve corresponding to the ROI at the passed index"""
        roi = self.__getitem__(ix)
        roi.set_color('w', width=2)

    def set_previous_index(self):
        """Set the previous_index attribute"""
        if self.previous_index == -1:
            self.previous_index = self.current_index
            return
        if self.previous_index > self.__len__() - 1:
            self.previous_index = self.current_index
            return

        roi = self.__getitem__(self.previous_index)
        self.previous_index = self.current_index
        roi.reset_color()

    def slot_show_all_checkbox_clicked(self, b: bool):
        """Show all ROIs in the viewer overlay visualization and curves"""
        if b:
            self._show_all_graphics_objects()
        else:
            self._hide_all_graphics_objects()
            ix = self.list_widget.currentRow()
            self._show_graphics_object(ix)

    def _show_graphics_object(self, ix: int):
        """Show the ROI at the passed index in the viewer overlay visualization"""
        try:
            roi = self.__getitem__(ix)
        except IndexError:
            return

        roi.curve_plot_item.show()
        roi_graphics_object = roi.get_roi_graphics_object()

        if roi_graphics_object is None:
            return

        roi_graphics_object.show()

    def _hide_graphics_object(self, ix: int):
        """Hide the ROI at the passed index in the viewer overlay visualization"""
        roi = self.__getitem__(ix)
        roi_graphics_object = roi.get_roi_graphics_object()
        roi.curve_plot_item.hide()

        if roi_graphics_object is None:
            return

        roi_graphics_object.hide()

    def _show_all_graphics_objects(self):
        """Show all ROIs in the viewer overlay visualization"""
        for ix in range(self.__len__()):
            self._show_graphics_object(ix)

    def _hide_all_graphics_objects(self):
        """Hide all ROIs in the viewer overlay visualization"""
        for ix in range(self.__len__()):
            self._hide_graphics_object(ix)

    def _live_update_requested(self, roi: Union[ManualROI, ScatterROI]):
        ix = self.index(roi)
        if not issubclass(self.roi_types, ManualROI):
            raise TypeError('Can only live update Manually drawn ROIs when they are moved')

        self.vi.workEnv_changed('ROI Region')

        if not self.live_plot_checkbox.isChecked():
            return

        self.set_pg_roi_plot(ix)

    def plot_manual_roi_regions(self):
        """Plot the ROI curves from the regions of all ManualROI instances in the list"""
        for ix in range(self.__len__()):
            roi = self.__getitem__(ix)
            pg_roi = roi.get_roi_graphics_object()
            self.set_pg_roi_plot(ix)
            roi.reset_color()
        if not self.show_all_checkbox.isChecked():
            self._hide_all_graphics_objects()
            self._show_graphics_object(self.current_index)

    def set_pg_roi_plot(self, ix: int):
        """Plot the ROI curve from the region of the ManualROI instance at the passed index"""
        image = self.vi.viewer.getProcessedImage()

        if image.ndim == 2:
            axes = (0, 1)
        elif image.ndim == 3:
            axes = (1, 2)
        else:
            return
        self.vi.viewer.status_bar_label.showMessage('Please wait, calculating intensity values for ROI: ' + str(ix))

        # Get the ROI region
        pg_roi = self.__getitem__(ix).get_roi_graphics_object()
        data = pg_roi.getArrayRegion((image.view(np.ndarray)), self.vi.viewer.imageItem, axes)

        if data is not None:
            data[data == 0] = np.nan

            if data.ndim == 2:
            # while data.ndim > 1:
                # data = data.mean(axis=1)
                y = np.nanmean(data)

            elif image.ndim == 3:
                # Set the curve
                y = np.nanmean(data, axis=(1, 2))

            roi = self.__getitem__(ix)
            roi.curve_plot_item.setData(y=y, x=self.vi.viewer.tVals)
            roi.set_color('w', width=2)
            roi.curve_plot_item.show()
        self.vi.viewer.status_bar_label.clearMessage()

    def slot_btn_set_tag(self):
        ix = self.current_index
        try:
            roi = self.__getitem__(ix)
        except IndexError:
            return

        roi_def = self.list_widget_tags.currentItem().text().split(': ')[0]
        tag = self.line_edit_tag.text()

        roi.set_tag(roi_def, tag)

        self.list_widget_tags.currentItem().setText(roi_def + ': ' + tag)
        self.line_edit_tag.clear()

        n = self.list_widget_tags.count() - 1
        i = self.list_widget_tags.currentRow() + 1
        self.list_widget_tags.setCurrentRow(min(i, n))

    def get_tag(self, ix: int, roi_def: str) -> str:
        roi_def = self.list_widget_tags.currentItem().text().split[0]
        roi = self.__getitem__(ix)
        return roi.get_tag(roi_def)

    def get_all_tags(self, ix: int) -> dict:
        roi = self.__getitem__(ix)
        return roi.get_all_tags()

    def set_list_widget_tags(self):
        """Set the tags list for the ROI at the current index"""
        self.list_widget_tags.clear()
        ix = self.current_index
        try:
            roi = self.__getitem__(ix)
        except IndexError:
            return

        tags_dict = roi.get_all_tags()

        roi_defs = sorted(tags_dict.keys())

        for roi_def in roi_defs:
            item = roi_def + ': ' + roi.get_tag(roi_def)
            self.list_widget_tags.addItem(item)

        if self.list_widget_tags.count() > 0:
            self.list_widget_tags.setCurrentRow(0)

    def update_roi_defs_from_configuration(self):
        """Update ROI_DEFs in the Tags list from the project configuration"""
        roi_defs = configuration.proj_cfg.options('ROI_DEFS')

        for roi in self:
            tags_dict = roi.get_all_tags()
            for roi_def in roi_defs:
                if roi_def not in tags_dict.keys():
                    roi.set_tag(roi_def, tag='untagged')
        self.set_list_widget_tags()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './control_widget.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_ControlWidget(object):
    def setupUi(self, ControlWidget):
        ControlWidget.setObjectName("ControlWidget")
        ControlWidget.resize(263, 680)
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(self.dockWidgetContents)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.listWidgetDataColumns = ListWidget(self.dockWidgetContents)
        self.listWidgetDataColumns.setObjectName("listWidgetDataColumns")
        self.verticalLayout.addWidget(self.listWidgetDataColumns)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_2 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout.addWidget(self.label_2)
        self.comboBoxGroupBy = ComboBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxGroupBy.sizePolicy().hasHeightForWidth())
        self.comboBoxGroupBy.setSizePolicy(sizePolicy)
        self.comboBoxGroupBy.setObjectName("comboBoxGroupBy")
        self.horizontalLayout.addWidget(self.comboBoxGroupBy)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.label_7 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_7.setObjectName("label_7")
        self.verticalLayout.addWidget(self.label_7)
        self.listWidgetColormap = ColormapListWidget(self.dockWidgetContents)
        self.listWidgetColormap.setObjectName("listWidgetColormap")
        self.verticalLayout.addWidget(self.listWidgetColormap)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_4 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_4.setObjectName("label_4")
        self.horizontalLayout_2.addWidget(self.label_4)
        self.comboBoxUUIDColumn = ComboBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxUUIDColumn.sizePolicy().hasHeightForWidth())
        self.comboBoxUUIDColumn.setSizePolicy(sizePolicy)
        self.comboBoxUUIDColumn.setObjectName("comboBoxUUIDColumn")
        self.horizontalLayout_2.addWidget(self.comboBoxUUIDColumn)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_9 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout_3.addWidget(self.label_9)
        self.comboBoxDPTCurve = ComboBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxDPTCurve.sizePolicy().hasHeightForWidth())
        self.comboBoxDPTCurve.setSizePolicy(sizePolicy)
        self.comboBoxDPTCurve.setObjectName("comboBoxDPTCurve")
        self.horizontalLayout_3.addWidget(self.comboBoxDPTCurve)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_5 = QtWidgets.QLabel(self.dockWidgetContents)
        self.label_5.setObjectName("label_5")
        self.horizontalLayout_5.addWidget(self.label_5)
        self.spinBoxSpotSize = QtWidgets.QSpinBox(self.dockWidgetContents)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.spinBoxSpotSize.sizePolicy().hasHeightForWidth())
        self.spinBoxSpotSize.setSizePolicy(sizePolicy)
        self.spinBoxSpotSize.setMinimum(1)
        self.spinBoxSpotSize.setMaximum(999)
        self.spinBoxSpotSize.setSingleStep(1)
        self.spinBoxSpotSize.setProperty("value", 10)
        self.spinBoxSpotSize.setObjectName("spinBoxSpotSize")
        self.horizontalLayout_5.addWidget(self.spinBoxSpotSize)
        self.verticalLayout.addLayout(self.horizontalLayout_5)
        self.checkBoxLiveUpdateFromFlowchart = QtWidgets.QCheckBox(self.dockWidgetContents)
        self.checkBoxLiveUpdateFromFlowchart.setChecked(False)
        self.checkBoxLiveUpdateFromFlowchart.setObjectName("checkBoxLiveUpdateFromFlowchart")
        self.verticalLayout.addWidget(self.checkBoxLiveUpdateFromFlowchart)
        self.pushButtonUpdatePlot = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButtonUpdatePlot.setObjectName("pushButtonUpdatePlot")
        self.verticalLayout.addWidget(self.pushButtonUpdatePlot)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.pushButtonSave = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButtonSave.setObjectName("pushButtonSave")
        self.horizontalLayout_4.addWidget(self.pushButtonSave)
        self.pushButtonLoad = QtWidgets.QPushButton(self.dockWidgetContents)
        self.pushButtonLoad.setObjectName("pushButtonLoad")
        self.horizontalLayout_4.addWidget(self.pushButtonLoad)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        ControlWidget.setWidget(self.dockWidgetContents)

        self.retranslateUi(ControlWidget)
        QtCore.QMetaObject.connectSlotsByName(ControlWidget)

    def retranslateUi(self, ControlWidget):
        _translate = QtCore.QCoreApplication.translate
        ControlWidget.setWindowTitle(_translate("ControlWidget", "Co&ntrols"))
        self.label.setText(_translate("ControlWidget", "Data Column(s)"))
        self.label_2.setText(_translate("ControlWidget", "Group by"))
        self.label_7.setText(_translate("ControlWidget", "Choose colormap:"))
        self.label_4.setText(_translate("ControlWidget", "UUID column"))
        self.label_9.setText(_translate("ControlWidget", "DPT Curve column"))
        self.label_5.setText(_translate("ControlWidget", "Spot size"))
        self.checkBoxLiveUpdateFromFlowchart.setText(_translate("ControlWidget", "Live update data from flowchart"))
        self.pushButtonUpdatePlot.setText(_translate("ControlWidget", "Update Plot"))
        self.pushButtonSave.setText(_translate("ControlWidget", "Save"))
        self.pushButtonLoad.setText(_translate("ControlWidget", "Load"))

from ....pyqtgraphCore.widgets.ComboBox import ComboBox
from ....pyqtgraphCore.widgets.ListWidget import ListWidget
from ...utils import ColormapListWidget



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @author: kushal

# Chatzigeorgiou Group
# Sars International Centre for Marine Molecular Biology

# GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007


import numpy as np
import pandas as pd
from multiprocessing import Pool
from tqdm import tqdm
from .data_types import Transmission
from ..common import get_sys_config
from uuid import uuid4
from typing import *
from PyQt5.QtWidgets import QMessageBox
import traceback


class StimulusExtraction:
    def __init__(self,
                 transmission: Transmission,
                 data_column: str,
                 stimulus_type: str,
                 start_offset: int,
                 end_offset: int,
                 zero_pos: str):

        self.t = transmission
        self.t.df.reset_index(drop=True, inplace=True)

        self.data_column = data_column
        self.stimulus_type = stimulus_type

        self.start_offset = start_offset
        self.end_offset = end_offset

        accepted = ['start_offset', 'stim_end', 'stim_center']
        if zero_pos not in accepted:
            raise ValueError(f'Invalid "zero_pos" argument. Must pass one of:\n{accepted}')

        self.zero_pos = zero_pos

    def extract(self) -> Transmission:

        samples = self.t.df.SampleID.unique()

        n_threads = get_sys_config()['_MESMERIZE_N_THREADS']

        with Pool(n_threads) as p:
            dfs = list(
                tqdm(
                    p.imap(self._per_sample, samples),
                    total=len(samples),
                    position=0,
                    desc='samples'
                )
            )

        bad_samples = [s for s in dfs if type(s) is tuple]

        bad_samples_ids = [s[0] for s in bad_samples]
        bad_sample_tbs = [s[1] for s in bad_samples]

        if len(bad_samples) > 0:
            mb = QMessageBox()
            mb.setIcon(QMessageBox.Warning)
            mb.setWindowTitle("Problematic Samples")
            mb.setText("There was an issue with the stimmaps in the following samples")
            mb.setInformativeText(f"Number of problematic samples: {len(bad_samples)}")
            mb.setDetailedText(  # prints the sample IDs and then traceback details for each
                '\n'.join(bad_samples_ids) + '\n\nDetailed Tracebacks:\n' +
                '\n'.join(
                    f"{bad_samples_ids[i]}\n\t{bad_sample_tbs[i]}" for i in range(len(bad_samples))
                )
            )
            mb.setStandardButtons(QMessageBox.Ok)
            mb.exec_()

        dfs = [d for d in dfs if type(d) is not tuple]

        self.t.df = pd.concat([df for df in dfs if df is not None]).reset_index(drop=True)

        # self.t.df = self.t.df.explode('_st_stim_curve')

        return self.t

    def _per_sample(self, sample_id: str) -> Union[pd.DataFrame, Tuple[str, str]]:
        try:
            sub_df = self.t.df[self.t.df.SampleID == sample_id].copy()
            sub_df.reset_index(drop=True, inplace=True)

            stim_df = sub_df.stim_maps.iloc[0][0][0][self.stimulus_type]

            stim_df['start'] = stim_df['start'].astype(np.int64)
            stim_df['end'] = stim_df['end'].astype(np.int64)

            stim_df = stim_df.sort_values(by='start').reset_index(drop=True)

        except Exception:
            return (sample_id, traceback.format_exc())

        out_dfs = []

        for ix, r in tqdm(stim_df.iterrows(), total=stim_df.index.size, position=1, desc='stimulus period'):
            sp_df = self._per_stimulus_period(sub_df, r['name'], r['start'], r['end'])
            out_dfs.append(sp_df)

        return pd.concat(out_dfs).reset_index(drop=True)

    # TODO: This can be done MUCH faster with DataFrame.explode but need to figure out how to keep all curve specific data
    def _per_stimulus_period(self, sub_df: pd.DataFrame, st_name: str, st_start: int, st_end: int) -> pd.DataFrame:
        """
        Extract a single stimulus period from all curves

        :param curves:
        :param st_name:
        :param st_start:
        :param st_end:
        :return:
        """

        curve_uuids = sub_df['uuid_curve'].values

        cols = sub_df.columns.to_list() + ['ST_NAME', 'ST_TYPE', '_ST_START_IX', '_ST_END_IX', '_ST_CURVE', 'ST_uuid']
        out_df = pd.DataFrame(columns=cols)

        for u in curve_uuids:
            # Do each separately to accommodate for varying curve lengths
            s = sub_df[sub_df['uuid_curve'] == u].iloc[0].copy()

            curve = s[self.data_column]

            start_ix, end_ix = self._apply_offsets(st_start, st_end, curve.size - 1)

            s['ST_NAME'] = st_name
            s['ST_TYPE'] = self.stimulus_type
            s['_ST_START_IX'] = start_ix
            s['_ST_END_IX'] = end_ix
            s['_ST_CURVE'] = curve[start_ix:end_ix]
            s['ST_uuid'] = str(uuid4())

            out_df = out_df.append(s, ignore_index=True)

        return out_df

    def _apply_offsets(self, start_ix, end_ix, max_ix) -> Tuple[int, int]:
        if self.zero_pos == 'start_offset':
            tstart = max(start_ix + self.start_offset, 0)
            tend = min(end_ix + self.end_offset, max_ix)

        elif self.zero_pos == 'stim_end':
            tstart = end_ix
            tend = min(end_ix + self.end_offset, max_ix)

        elif self.zero_pos == 'stim_center':
            tstart = int(((start_ix + end_ix) / 2)) + self.start_offset
            tend = min(end_ix + self.end_offset, max_ix)

        return tstart, tend



import isx
import numpy as np
import click
from multiprocessing import shared_memory
from tqdm import tqdm


@click.command()
@click.option('--isx-path', type=str)
@click.option('--shm-meta-array-name', type=str)
def load_file(isx_path, shm_meta_array_name):#shm_name, shm-size, shm-dtype):
    """
    isx_path: full path to the isx file to open
    shm_meta_array_name: same of the shared memory array that is used to pass
    array metadata (shape, dtype, name) to effectively communication with the
    parent mesmerize instance

    shm-name: name of the parent shared array that contains the 

    """
    movie = isx.Movie.read(isx_path)

    meta = movie.get_acquisition_info()

    d = \
        {
            'fps': 1000 / movie.timing.period.to_msecs(),
            'origin': meta['Microscope Type'],
            'date': '00000000_000000',
            'orig_meta': meta
        }
    
    # total number of frames in the recording
    nframes = movie.timing.num_samples

    # shape of each frame
    frame0 = movie.get_frame_data(0)
    
    # image shape
    shape = (*frame0.shape, nframes)

    # calculate number of bytes required for the shared memory
    nbytes = np.prod(shape) * np.dtype(movie.data_type).itemsize

    # created a shared memory buffer
    shm = shared_memory.SharedMemory(create=True, size=nbytes)

    # create shared numpy array backed by the shared memory buffer
    imgseq = np.ndarray(shape, dtype=np.dtype(movie.data_type), buffer=shm.buf)
    
    # fill the array with the inscopix data
    for i in tqdm(range(nframes)):
        imgseq[..., i] = movie.get_frame_data(i)

    # communicate the array metadata to access the shared array from within mesmerize
    shm_metadata = shared_memory.SharedMemory(name=shm_meta_array_name)
    array_metadata = np.ndarray((5,), dtype=np.dtype('<U64'), buffer=shm_metadata.buf)

    array_metadata[0] = shm.name
    array_metadata[1] = np.dtype(movie.data_type).name
    array_metadata[2] = ','.join(str(s) for s in shape)
    array_metadata[3] = str(d['fps'])
    array_metadata[4] = str(d['origin'])
    
    print(f'imgdata address: {shm.name}')
    print(array_metadata)

    shm.close()


if __name__ == '__main__':
    load_file()





from .variants import *
from .widgets import *

from ..analysis import Transmission
from ..common import get_project_manager
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import Qt
from os.path import basename


def open_plot_file(file_path: str) -> QWidget:
    t = Transmission.from_hdf5(file_path)

    if t.plot_state is None:
        raise ValueError('The chosen Transmission file does not have the required "plot_state" attribute')

    if t.plot_state['type'] not in widgets.__all__:
        raise TypeError(f'Plot type not found: {t.plot_state["type"]}.\nSupported plot types are:\n{widgets.__all__}')

    plot_widget = getattr(widgets, t.plot_state['type'])()
    plot_widget.open_plot(file_path, get_project_manager().root_dir)
    plot_widget.setAttribute(Qt.WA_DeleteOnClose)
    plot_widget.setWindowTitle(f"{plot_widget.windowTitle()} - {basename(file_path)}")

    return plot_widget



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './controls.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_DockWidget(object):
    def setupUi(self, DockWidget):
        DockWidget.setObjectName("DockWidget")
        DockWidget.resize(267, 690)
        DockWidget.setMinimumSize(QtCore.QSize(96, 690))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        DockWidget.setFont(font)
        self.dockWidgetContents = QtWidgets.QWidget()
        self.dockWidgetContents.setObjectName("dockWidgetContents")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.toolBox = QtWidgets.QToolBox(self.dockWidgetContents)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.toolBox.setFont(font)
        self.toolBox.setObjectName("toolBox")
        self.page = QtWidgets.QWidget()
        self.page.setGeometry(QtCore.QRect(0, 0, 251, 578))
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.page.setFont(font)
        self.page.setObjectName("page")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.page)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout()
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(self.page)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.comboBoxDataColumn = QtWidgets.QComboBox(self.page)
        self.comboBoxDataColumn.setObjectName("comboBoxDataColumn")
        self.horizontalLayout.addWidget(self.comboBoxDataColumn)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout_3.addLayout(self.horizontalLayout)
        self.label_2 = QtWidgets.QLabel(self.page)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_3.addWidget(self.label_2)
        self.comboBoxStimulusMapping = QtWidgets.QComboBox(self.page)
        self.comboBoxStimulusMapping.setObjectName("comboBoxStimulusMapping")
        self.verticalLayout_3.addWidget(self.comboBoxStimulusMapping)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.label_3 = QtWidgets.QLabel(self.page)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_2.addWidget(self.label_3)
        self.comboBoxMethod = QtWidgets.QComboBox(self.page)
        self.comboBoxMethod.setObjectName("comboBoxMethod")
        self.horizontalLayout_2.addWidget(self.comboBoxMethod)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.verticalLayout_3.addLayout(self.horizontalLayout_2)
        self.groupBox = QtWidgets.QGroupBox(self.page)
        self.groupBox.setToolTip("")
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.plainTextEditMethodParams = QtWidgets.QPlainTextEdit(self.groupBox)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.plainTextEditMethodParams.setFont(font)
        self.plainTextEditMethodParams.setObjectName("plainTextEditMethodParams")
        self.verticalLayout.addWidget(self.plainTextEditMethodParams)
        self.verticalLayout_3.addWidget(self.groupBox)
        self.checkBoxUsePreprocessScaler = QtWidgets.QCheckBox(self.page)
        self.checkBoxUsePreprocessScaler.setChecked(True)
        self.checkBoxUsePreprocessScaler.setObjectName("checkBoxUsePreprocessScaler")
        self.verticalLayout_3.addWidget(self.checkBoxUsePreprocessScaler)
        self.comboBoxPreprocessScaler = QtWidgets.QComboBox(self.page)
        self.comboBoxPreprocessScaler.setObjectName("comboBoxPreprocessScaler")
        self.verticalLayout_3.addWidget(self.comboBoxPreprocessScaler)
        self.label_5 = QtWidgets.QLabel(self.page)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_3.addWidget(self.label_5)
        self.listWidgetStimulusColormap = ColormapListWidget(self.page)
        self.listWidgetStimulusColormap.setObjectName("listWidgetStimulusColormap")
        self.verticalLayout_3.addWidget(self.listWidgetStimulusColormap)
        self.pushButtonComputeAllSamples = QtWidgets.QPushButton(self.page)
        self.pushButtonComputeAllSamples.setObjectName("pushButtonComputeAllSamples")
        self.verticalLayout_3.addWidget(self.pushButtonComputeAllSamples)
        self.verticalLayout_6.addLayout(self.verticalLayout_3)
        self.progressBar = QtWidgets.QProgressBar(self.page)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setObjectName("progressBar")
        self.verticalLayout_6.addWidget(self.progressBar)
        self.toolBox.addItem(self.page, "")
        self.page_2 = QtWidgets.QWidget()
        self.page_2.setGeometry(QtCore.QRect(0, 0, 251, 578))
        self.page_2.setObjectName("page_2")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.page_2)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.pushButton_3 = QtWidgets.QPushButton(self.page_2)
        self.pushButton_3.setObjectName("pushButton_3")
        self.verticalLayout_5.addWidget(self.pushButton_3)
        self.listWidgetComputedSamples = QtWidgets.QListWidget(self.page_2)
        font = QtGui.QFont()
        font.setBold(False)
        font.setWeight(50)
        self.listWidgetComputedSamples.setFont(font)
        self.listWidgetComputedSamples.setObjectName("listWidgetComputedSamples")
        self.verticalLayout_5.addWidget(self.listWidgetComputedSamples)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.pushButton = QtWidgets.QPushButton(self.page_2)
        self.pushButton.setObjectName("pushButton")
        self.horizontalLayout_3.addWidget(self.pushButton)
        self.verticalLayout_5.addLayout(self.horizontalLayout_3)
        self.toolBox.addItem(self.page_2, "")
        self.verticalLayout_2.addWidget(self.toolBox)
        DockWidget.setWidget(self.dockWidgetContents)

        self.retranslateUi(DockWidget)
        self.toolBox.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(DockWidget)
        DockWidget.setTabOrder(self.comboBoxDataColumn, self.comboBoxStimulusMapping)
        DockWidget.setTabOrder(self.comboBoxStimulusMapping, self.comboBoxMethod)
        DockWidget.setTabOrder(self.comboBoxMethod, self.plainTextEditMethodParams)
        DockWidget.setTabOrder(self.plainTextEditMethodParams, self.checkBoxUsePreprocessScaler)
        DockWidget.setTabOrder(self.checkBoxUsePreprocessScaler, self.comboBoxPreprocessScaler)
        DockWidget.setTabOrder(self.comboBoxPreprocessScaler, self.listWidgetStimulusColormap)
        DockWidget.setTabOrder(self.listWidgetStimulusColormap, self.pushButtonComputeAllSamples)
        DockWidget.setTabOrder(self.pushButtonComputeAllSamples, self.pushButton_3)
        DockWidget.setTabOrder(self.pushButton_3, self.listWidgetComputedSamples)
        DockWidget.setTabOrder(self.listWidgetComputedSamples, self.pushButton)

    def retranslateUi(self, DockWidget):
        _translate = QtCore.QCoreApplication.translate
        DockWidget.setWindowTitle(_translate("DockWidget", "Co&ntrols"))
        self.label.setText(_translate("DockWidget", "Data Column:"))
        self.label_2.setText(_translate("DockWidget", "Stimulus/Behavior mapping:"))
        self.label_3.setText(_translate("DockWidget", "Method:"))
        self.groupBox.setTitle(_translate("DockWidget", "Params for method"))
        self.plainTextEditMethodParams.setPlainText(_translate("DockWidget", "n_components=2"))
        self.checkBoxUsePreprocessScaler.setText(_translate("DockWidget", "Use preprocess scaler:"))
        self.label_5.setText(_translate("DockWidget", "Stimulus colormap:"))
        self.pushButtonComputeAllSamples.setText(_translate("DockWidget", "Compute for all samples"))
        self.toolBox.setItemText(self.toolBox.indexOf(self.page), _translate("DockWidget", "Parameters"))
        self.pushButton_3.setText(_translate("DockWidget", "Open Selected Sample In Viewer"))
        self.pushButton.setText(_translate("DockWidget", "Save"))
        self.toolBox.setItemText(self.toolBox.indexOf(self.page_2), _translate("DockWidget", "Computed Samples"))

from ...utils import ColormapListWidget



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

import sys
import numpy as np
import pickle
import os
from mesmerize.common.configuration import HAS_TSLEARN, get_sys_config
if HAS_TSLEARN:
    from tslearn.clustering import KShape
from joblib import Parallel, delayed
from typing import *
from scipy import signal
from itertools import product
import json

def run(data_path: str, params_path: str):
    X = np.load(data_path)

    params = pickle.load(open(params_path, 'rb'))
    workdir = params['workdir']

    out = os.path.join(workdir, 'out')
    with open(out, 'w') as f:
        f.write('0')

    print(f'Using work dir: {workdir}')

    print('** Fitting training data **')
    n_train = int((params['kwargs'].pop('train_percent') / 100) * X.shape[0])
    train = X[np.random.choice(X.shape[0], size=n_train, replace=False)]

    if params['do_single']:
        run_single(X, train, params['kwargs'], workdir, out)
    else:
        run_grid(X, train, params['kwargs'], workdir, out)


def run_single(X, train, params, workdir, out):
    kwargs = params
    ks = KShape(**kwargs)

    ks.fit(train)

    print('**** Predicting ****')
    y_pred = ks.predict(X)

    ks_path = os.path.join(workdir, 'ks.pickle')
    pickle.dump(ks, open(ks_path, 'wb'))

    y_pred_path = os.path.join(workdir, 'y_pred.npy')
    np.save(y_pred_path, y_pred)

    train_path = os.path.join(workdir, 'train.npy')
    np.save(train_path, train)

    with open(out, 'w') as f:
        f.write('1')

    print('* Done! *')


# np.array_split(a, n_partitions, axis=0)

def sort_peak_widths(a: np.ndarray):
    widths = [signal.peak_widths(p, [np.argmax(p)])[0][0] for p in a]
    sort_ixs = np.argsort(widths)

    return a[sort_ixs]


def sort_amplitidue(a: np.ndarray):
    ampls = [s.argmax() for s in a]
    sort_ixs = np.argsort(ampls)

    return a[sort_ixs]


def kshape_grid_iter(X_partitioned: List[np.array], kshape_kwargs: dict) -> Tuple[KShape, int]:
    seed_ixs = [np.random.randint(0, X.shape[0] - 1) for X in X_partitioned]
    centroid_seeds = np.array([X_partitioned[i][seed] for i, seed in enumerate(seed_ixs)])
    init = np.swapaxes(np.array([centroid_seeds]).T, 0, 1)

    kshape = KShape(
        n_clusters=len(seed_ixs),
        init=init,
        verbose=True,
        random_state=None,
        **kshape_kwargs
    )

    X = np.vstack(X_partitioned)

    print('** Fitting ks model **')
    kshape.fit(X)

    print('** Predicting **')
    n_clusters_out = np.unique(kshape.predict(X)).size

    # until the tslearn hyper-param json issue is released in the latest pypi version
    kshape.init = kshape.init.tolist()

    return kshape, n_clusters_out


def run_grid(X, train, params, workdir, out):
    sortby = params.pop('sortby')
    if sortby == 'width':
        X_sorted = sort_peak_widths(X)
    elif sortby == 'amplitude':
        X_sorted = sort_amplitidue(X)

    p_range = params.pop('npartitions_range')
    ncombinations = params.pop('ncombinations')

    n_jobs = get_sys_config()['_MESMERIZE_N_THREADS']
    tmpdir = os.path.join(workdir, 'joblib_tmp')

    kg = Parallel(n_jobs=n_jobs, verbose=5, temp_folder=tmpdir)(
        delayed(kshape_grid_iter)(
            np.array_split(X_sorted, nparts, axis=0), params
            ) for nparts in range(*p_range) \
                for i in range(ncombinations)
    )

    kga = [k[0]._to_dict(output='json') for k in kg]

    n_clusters_out = np.array([k[1] for k in kg]).reshape(
        len(range(*p_range)),
        ncombinations
    )

    json.dump(kga, open(os.path.join(workdir, 'kga.json'), 'w'))
    np.save(os.path.join(workdir, 'n_clusters_out.npy'), n_clusters_out)

    with open(out, 'w') as f:
        f.write('1')

    print('* Done! *')


#if sys.argv[0] == __file__:
if __name__ == '__main__':
    run(sys.argv[1], sys.argv[2])



import sys
from ...common.configuration import HAS_CAIMAN, HAS_TENSORFLOW#, HAS_ISX

__all__ = [
    'mesfile_io',
    'roi_manager',
    'tiff_io',
    'stimulus_mapping',
    'script_editor',
    'suite2p',
    'femtonics_mesc',
    'exporter',
]

if HAS_CAIMAN:
    __all__ += \
    [
        'cnmfe',
        'cnmf',
        'cnmf_3d',
        'caiman_motion_correction',
        'caiman_importer',
        'caiman_dfof'
    ]

if HAS_TENSORFLOW:
    __all__ += \
        [
            'nuset_segment',
        ]

if sys.version_info[1] >= 8:
    __all__ += ['inscopix_importer']

#if HAS_ISX:
#    __all__ += \
#    [
#        'inscopix_importer'
#    ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May 13 2018

@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from ..core.common import ViewerUtils
from .pytemplates.caiman_motion_correction_pytemplate import *
from ...common import get_window_manager
from ...pyqtgraphCore import LinearRegionItem
from uuid import UUID
from copy import deepcopy
from joblib import Parallel, delayed
import numpy as np
from psutil import cpu_count
import tifffile
from ...pyqtgraphCore.widgets.MatplotlibWidget import MatplotlibWidget
import os
from math import ceil
from itertools import product
from typing import *
import logging


logger = logging.getLogger()


class ModuleGUI(QtWidgets.QDockWidget):
    def __init__(self, parent, viewer_reference):
        self.vi = ViewerUtils(viewer_reference)

        QtWidgets.QDockWidget.__init__(self, parent)

        self.ui = Ui_DockWidget()
        self.ui.setupUi(self)

        self.ui.btnAddToBatchElastic.clicked.connect(self.add_to_batch_elas_corr)

        self.ui.btnShowQuilt.clicked.connect(self.draw_quilt)

        self.ui.sliderOverlaps.valueChanged.connect(self.draw_quilt)
        self.ui.sliderStrides.valueChanged.connect(self.draw_quilt)

        self.overlapsH = []
        self.overlapsV = []

        self.ncols_projection_plots = 3
        self.cmap_projection_plots = 'gray'
        self.projections_plot_widget = None
        self.projections_plot_vmin_vmax_stds = 5

        self.currrent_projections: List[np.ndarray] = None

        self.ui.pushButtonViewProjections.clicked.connect(self.view_output_projections)

    def draw_quilt(self):
        if self.ui.btnShowQuilt.isChecked() is False:
            self.remove_quilt()
            return
        if len(self.overlapsV) > 0:
            for overlap in self.overlapsV:
                self.vi.viewer.view.removeItem(overlap)
            for overlap in self.overlapsH:
                self.vi.viewer.view.removeItem(overlap)
            self.overlapsV = []
            self.overlapsH = []

        w = int(self.vi.viewer.view.addedItems[0].width())
        k = self.ui.sliderStrides.value()

        h = int(self.vi.viewer.view.addedItems[0].height())
        j = self.ui.sliderStrides.value()

        val = int(self.ui.sliderOverlaps.value())

        for i in range(1, int(w / k) + 1):
            linreg = LinearRegionItem(values=[i * k, i * k + val], brush=(255, 255, 255, 80),
                                      movable=False, bounds=[i * k, i * k + val])
            self.overlapsV.append(linreg)
            self.vi.viewer.view.addItem(linreg)

        for i in range(1, int(h / j) + 1):
            linreg = LinearRegionItem(values=[i * j, i * j + val], brush=(255, 255, 255, 80),
                                      movable=False, bounds=[i * j, i * j + val],
                                      orientation=LinearRegionItem.Horizontal)
            self.overlapsH.append(linreg)
            self.vi.viewer.view.addItem(linreg)

    def remove_quilt(self):
        for o in self.overlapsV:
            self.vi.viewer.view.removeItem(o)
        for o in self.overlapsH:
            self.vi.viewer.view.removeItem(o)
        self.overlapsH = []
        self.overlapsV = []

    def get_params(self, group_params: bool = False) -> dict:
        """
        Get a dict of the set parameters
        :return: parameters dict
        :rtype: dict
        """

        if self.ui.spinBoxGSig_filt.value() == 0:
            gSig_filt = None
        else:
            gSig = self.ui.spinBoxGSig_filt.value()
            gSig_filt = (gSig, gSig)

        d = \
            {
                'output_bit_depth': self.ui.comboBoxOutputBitDepth.currentText(),
                'template': None
            }

        mc_params = \
            {
                'max_shifts': (self.ui.spinboxX.value(), self.ui.spinboxY.value()),
                'niter_rig': self.ui.spinboxIterRigid.value(),
                'max_deviation_rigid': self.ui.spinboxMaxDev.value(),
                'strides': (self.ui.sliderStrides.value(), self.ui.sliderStrides.value()),
                'overlaps': (self.ui.sliderOverlaps.value(), self.ui.sliderOverlaps.value()),
                'upsample_factor_grid': self.ui.spinboxUpsample.value(),
                'gSig_filt': gSig_filt
            }

        # Update the dict with any user entered kwargs
        if self.ui.groupBox_motion_correction_kwargs.isChecked():
            _kwargs = self.ui.plainTextEdit_mc_kwargs.toPlainText()
            mc_params.update(eval(f"dict({_kwargs})"))

        if self.vi.viewer.workEnv.imgdata.ndim == 4:
            is_3d = True
        else:
            is_3d = False

        d['is_3d'] = is_3d

        if group_params:
            d.update({'mc_kwargs': mc_params})

        else:
            d.update({**mc_params})

        d['keep_memmap'] = self.ui.checkBoxKeepMemmap.isChecked()

        return d

    def add_to_batch_elas_corr(self):
        """
        Add a batch item with the currently set parameters and the current work environment.
        """

        name = self.ui.lineEditNameElastic.text()

        self.vi.viewer.status_bar_label.showMessage(
            'Please wait, adding CaImAn motion correction: ' + name + ' to batch...')

        self.add_to_batch()

        self.vi.viewer.status_bar_label.showMessage('Done adding CaImAn motion correction: ' + name + ' to batch!')

        self.ui.lineEditNameElastic.clear()

    def add_to_batch(self, params: dict = None) -> UUID:
        if params is None:
            input_params = self.get_params(group_params=True)
            item_name = self.ui.lineEditNameElastic.text()
            input_params['item_name'] = item_name
        else:
            # check that params dict is formatted properly
            required_keys = ['item_name', 'mc_kwargs', 'output_bit_depth']
            if any([k not in params.keys() for k in required_keys]):
                raise ValueError(f'Must pass a params dict with the following keys:\n'
                                 f'{required_keys}\n'
                                 f'Please see the docs for more information.')
            input_params = deepcopy(params)
            item_name = input_params['item_name']

        input_workEnv = self.vi.viewer.workEnv
        batch_manager = get_window_manager().get_batch_manager()
        u = batch_manager.add_item(module='caiman_motion_correction',
                                   name=item_name,
                                   input_workEnv=input_workEnv,
                                   input_params=input_params,
                                   info=input_params
                                   )

        return u

    @staticmethod
    def _get_projection(proj_type: str,  batch_path: str, item_uuid: UUID):
        tiff_path = os.path.join(batch_path, f'{item_uuid}_mc.tiff')
        tif = tifffile.TiffFile(tiff_path, is_nih=True)
        seq = tif.asarray(maxworkers=cpu_count())

        func = getattr(np, f"nan{proj_type}")
        return func(seq, axis=0)

    def view_output_projections(self):
        batch_manager = get_window_manager().get_batch_manager()
        batch_path = batch_manager.batch_path

        items = batch_manager.get_selected_items()  # get a list of indices and UUIDs for the selected items
        self._projection_item_indices = items[0]
        item_uuids = items[1]

        if len(self._projection_item_indices) > 3:
            if QtWidgets.QMessageBox.question(
                self, 'Many items selected',
                f"You have selected < {len(self._projection_item_indices)} > items, "
                f"it may take a while to calculate these projections.\n"
                f"Do you still want to continue?",
                QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.Yes
            ) == QtWidgets.QMessageBox.No:
                return

        proj = self.ui.comboBoxProjectionsOption.currentText()

        logger.info("Calculating projections, please wait...")

        self.currrent_projections = Parallel(n_jobs=cpu_count(), verbose=5)(
            delayed(ModuleGUI._get_projection)(proj, batch_path, item_uuid) for item_uuid in item_uuids
        )

        self.nrows_projection_plots = ceil(len(self._projection_item_indices) / self.ncols_projection_plots)
        self.ncols_projection_plots = min(len(self._projection_item_indices), 3)

        self.plot_output_projections()

    def plot_output_projections(self):
        if self.currrent_projections is None:
            raise AttributeError("Projections not calculated")

        logger.info("Plotting projections")
        if self.projections_plot_widget is None:
            self.projections_plot_widget = MatplotlibWidget()
        else:
            self.projections_plot_widget.fig.clear()

        self.projections_plot_widget.axs = self.projections_plot_widget.fig.subplots(self.nrows_projection_plots, self.ncols_projection_plots)

        for i, axes_ix in enumerate(
                product(range(self.nrows_projection_plots), range(self.ncols_projection_plots))
        ):
            if not i < len(self._projection_item_indices):
                break

            if self.nrows_projection_plots == 1:
                axes_ix = axes_ix[1]

            index = self._projection_item_indices[i]
            projection = self.currrent_projections[i]

            proj_mean = np.nanmean(projection)
            proj_std = np.nanstd(projection)

            z = self.projections_plot_vmin_vmax_stds * proj_std

            vmin = max(proj_mean - z, np.nanmin(projection))
            vmax = min(proj_mean + z, np.nanmax(projection))

            self.projections_plot_widget.axs[axes_ix].imshow(
                projection,
                cmap=self.cmap_projection_plots,
                vmin=vmin,
                vmax=vmax
            )
            self.projections_plot_widget.axs[axes_ix].set_title(index)
            self.projections_plot_widget.axs[axes_ix].set_title(index)

        self.projections_plot_widget.draw()
        self.projections_plot_widget.fig.tight_layout()
        self.projections_plot_widget.show()



"""
Subclass of QLineEdit because for whatever reason
the returnPressed signal causes the entire application
to crash when used in flowchart parameter widgets.
"""


from PyQt5 import QtWidgets


class DummySignal:
    def connect(self, *args, **kwargs):
        pass


class LineEdit(QtWidgets.QLineEdit):
    def returnPressed(self, *args, **kwargs):
        return DummySignal()

    def saveState(self):
        return self.text()

    def restoreState(self, txt):
        self.setText(txt)

    def widgetGroupInterface(self):
        return (self.returnPressed, self.saveState, self.restoreState)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: kushal

Chatzigeorgiou Group
Sars International Centre for Marine Molecular Biology

GNU GENERAL PUBLIC LICENSE Version 3, 29 June 2007
"""

from PyQt5 import QtWidgets, QtCore
from mesmerize import Transmission
from ..base import BasePlotWidget
from ....common.qdialogs import exceptions_label
from ....common.configuration import console_history_path
from ....pyqtgraphCore.console import ConsoleWidget
from typing import *
from uuid import UUID
from .control_widget import Ui_ControlWidget
from ..datapoint_tracer import DatapointTracerWidget
import os


class ControlDock(QtWidgets.QDockWidget):
    sig_changed = QtCore.pyqtSignal()

    def __init__(self, parent):
        QtWidgets.QDockWidget.__init__(self, parent=parent)
        self.ui = Ui_ControlWidget()
        self.ui.setupUi(self)

        self.ui.listWidgetColormap.set_cmap('tab10')

        self.ui.listWidgetDataColumns.selectionChanged.connect(self.sig_changed.emit)
        self.ui.comboBoxGroupBy.currentTextChanged.connect(self.sig_changed.emit)
        self.ui.listWidgetColormap.currentRowChanged.connect(self.sig_changed.emit)
        self.ui.comboBoxUUIDColumn.currentTextChanged.connect(self.sig_changed.emit)
        self.ui.comboBoxDPTCurve.currentTextChanged.connect(self.sig_changed.emit)
        self.ui.spinBoxSpotSize.valueChanged.connect(self.sig_changed.emit)

        self.data_controls = [self.ui.listWidgetDataColumns, self.ui.comboBoxDPTCurve]
        self.categorical_controls = [self.ui.comboBoxGroupBy]
        self.uuid_controls = [self.ui.comboBoxUUIDColumn]

    def fill_widget(self, data_columns: list, categorical_columns: list, uuid_columns: list):
        for w in self.data_controls:
            w.setItems(data_columns)

        for w in self.categorical_controls:
            w.setItems(categorical_columns)

        for w in self.uuid_controls:
            w.setItems(uuid_columns)

    def set_combo_box(self, comboBox: QtWidgets.QComboBox, text: str):
        ix = comboBox.findText(text)
        comboBox.setCurrentIndex(ix)


class CentralWidget(QtWidgets.QWidget):
    def __init__(self, parent):
        QtWidgets.QWidget.__init__(self, parent=parent)
        self.status_label = QtWidgets.QLabel(self)
        self.status_label.setMaximumHeight(32)
        self.vlayout = QtWidgets.QVBoxLayout()
        self.vlayout.addWidget(self.status_label)


class BeeswarmPlotWidget(QtWidgets.QMainWindow, BasePlotWidget):
    drop_opts = []

    def __init__(self):
        QtWidgets.QMainWindow.__init__(self, parent=None)
        BasePlotWidget.__init__(self)
        self.setWindowTitle('Beeswarm Plot')

        self.central_widget = CentralWidget(self)
        self.status_label = self.central_widget.status_label

        self.exception_holder = None

        self.control_widget = ControlDock(self)
        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.control_widget)
        # self.control_widget.sig_changed.connect(self.update_plot)

        # Update only when button is clicked since these plots can take a while to draw.
        self.control_widget.ui.pushButtonUpdatePlot.clicked.connect(self.update_plot)
        self.control_widget.ui.checkBoxLiveUpdateFromFlowchart.toggled.connect(self.set_update_live)

        self.block_signals_list = [self.control_widget]

        self.transmision_withhold = None

        self.live_datapoint_tracer = DatapointTracerWidget()
        self.dockDatapointTracer = QtWidgets.QDockWidget(self)
        self.dockDatapointTracer.setWindowTitle('Datapoint Tracer')
        self.dockDatapointTracer.setFeatures(QtWidgets.QDockWidget.DockWidgetFloatable | QtWidgets.QDockWidget.DockWidgetMovable)
        self.dockDatapointTracer.setWidget(self.live_datapoint_tracer)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.dockDatapointTracer)

        cmd_history_file = os.path.join(console_history_path, 'beeswarm_plot.pik')

    def set_update_live(self, b: bool):
        self.update_live = b
        self.control_widget.ui.checkBoxLiveUpdateFromFlowchart.setChecked(b)

        if b and (self.transmision_withhold is not None):
            self.set_input(self.transmision_withhold)  #: set the input that has been withheld

    @BasePlotWidget.signal_blocker
    def set_input(self, transmission: Transmission):
        if (self.transmission is None) or self.update_live:
            super(BeeswarmPlotWidget, self).set_input(transmission)
            return

        else:  # hold the transmission in case the user clicks the update checkbox
            self.transmision_withhold = transmission

    def get_plot_opts(self, drop: bool = False) -> dict:
        d = {'data_columns': self.control_widget.ui.listWidgetDataColumns.getSelectedItems(),
             'group_by_column': self.control_widget.ui.comboBoxGroupBy.currentText(),
             'cmap': self.control_widget.ui.listWidgetColormap.current_cmap,
             'uuid_column': self.control_widget.ui.comboBoxUUIDColumn.currentText(),
             'dpt_curve_column': self.control_widget.ui.comboBoxDPTCurve.currentText(),
             'spot_size': self.control_widget.ui.spinBoxSpotSize.value()
             }

        return d

    @BasePlotWidget.signal_blocker
    def set_plot_opts(self, opts: dict):
        self.control_widget.ui.listWidgetDataColumns.setSelectedItems(opts['data_columns'])
        self.control_widget.set_combo_box(self.control_widget.ui.comboBoxGroupBy, opts['group_by_column'])
        self.control_widget.ui.listWidgetColormap.set_cmap(opts['cmap'])
        self.control_widget.set_combo_box(self.control_widget.ui.comboBoxUUIDColumn, opts['uuid_column'])
        self.control_widget.set_combo_box(self.control_widget.ui.comboBoxDPTCurve, opts['dpt_curve_column'])
        self.control_widget.ui.spinBoxSpotSize.setValue(opts['spot_size'])

    @exceptions_label('status_label', 'exception_holder', 'Error while setting data', 'Make sure you have selected approriate columns')
    def update_plot(self, *args, **kwargs):
        opts = self.get_plot_opts()
        self.plot_opts = opts

        data_columns = opts['data_columnns']

        # for col in data_columns:
        #     data =

    def set_current_datapoint(self, identifier: Union[UUID, str]):
        self.current_datapoint = identifier
        u = str(self.current_datapoint)

    def show_exception_info(self, mouse_ev):
        if self.exception_holder is not None:
            QtWidgets.QMessageBox.warning(self, *self.exception_holder)